#include "Slicer.h"
#include <string.h>

static CCB MyCCB = {		/* This is for saving 3DO shapes */
	0x47664010,		/* 0x20 = No Mask, 0x200 = Packed */
	0x00000000,		/* ccb_NextPtr */
	0x00000030,	/* ccb_CelData */
	0x00000000,	/* ccb_PIPPtr */
	0x00000000,	/* ccb_X */
	0x00000000,	/* ccb_Y */
	0x00100000,	/* ccb_HDX */
	0x00000000,	/* ccb_HDY */
	0x00000000,	/* ccb_VDX */
	0x00010000,	/* ccb_VDY */
	0x00000000,	/* ccb_DDX */
	0x00000000,	/* ccb_DDY */
	0x1F001F00,	/* ccb_PIXC */
	0x00000016,	/* ccb_PRE0 */
	0x03FF1000	/* ccb_PRE1 Low bit is X control*/
};


Word ColorBits3DO=16;			/* Bits per pixel to save */
static Word SkipLineSize;		/* Number of bytes for scan line padding */
static Word16 Pal3DO[256];		/* Palette in 3DO colors */
static Word8 Translate[256];		/* Color translate table */
static Word8 ColorUsed[256];		/* Which colors are used? */
static Word ColorCount;			/* Number of colors in PLUT */
static Word16 PLUT[32];			/* Save palette lookup table */
static Word16 One3DOLine[1000];		/* One scan line of 3DO data */
static Word8 PackedLine[2000];		/* One compressed line of 3DO data */
static Word PIndex;					/* Index to compressed data */
static Word8 PBitBucket;				/* Dest bit bucket */
static Word8 PBitCount;				/* Dest bit count */

/***************************************

	Convert the 256 color palette into a format for the 3DO

***************************************/

static void Make3DOPal(Word8 *Palette)
{
	Word i,j;		/* Temps */
	Word Var;		/* Temp color */
	Word Red,Green,Blue;		/* Color components */

	i = 0;
	j = 0;
	do {
		Red = Palette[j]&0xF8;	/* Get red */
		Green = Palette[j+1]&0xF8;	/* Get green */
		Blue = Palette[j+2]&0xF8;	/* Get blue */
		Var = (Red<<7)|(Green<<2)|(Blue>>3);	/* Blend together */
		Pal3DO[i] = static_cast<Word16>(Var);		/* Save the 15 bit color version */
		j+=3;				/* Next color */
	} while (++i<256);		/* All done? */
}


/***************************************

	Save some bits into the data stream

***************************************/

static void WritePackBits(Word Input,Word BitCount)
{
	Word Reversed;
	Word i;

	i = 0;
	Reversed = 0;		/* Init the bits reversed */
	do {
		Reversed<<=1;	/* Shift up result */
		if (Input&1) {	/* Is the low bit set? */
			Reversed|=1;	/* Set in result */
		}
		Input>>=1;		/* Discard the bit */
	} while (++i<BitCount);	/* Done? */
	i = 0;
	do {
		if (PBitCount==8) {		/* Bit bucket full? */
			PackedLine[PIndex] = PBitBucket;	/* Save the result */
			++PIndex;
			PBitCount = 0;		/* Reset the count */
		}
		PBitBucket<<=1;			/* Shift up the bucket */
		if (Reversed&1) {		/* Should I save the bit? */
			PBitBucket|=1;		/* Send it off */
		}
		Reversed>>=1;			/* Shift it out */
		++PBitCount;			/* Adjust the count */
	} while (++i<BitCount);		/* All done? */
}

/***************************************

	Flush the packbits buffer

***************************************/

static void FlushPackBits(void)
{
	if (PBitCount) {
		while (PBitCount!=8) {		/* Full yet (Pad it out) */
			++PBitCount;
			PBitBucket<<=1;			/* Shift up */
		}
		PackedLine[PIndex] = PBitBucket;	/* Save the final byte */
		++PIndex;			/* Next dest index */
		PBitCount = 0;		/* No bits in buffer */
	}
}

/***************************************

	Write out a line of UNCOMPRESSED shape data

***************************************/

static void WriteUnpacked(Word Width,FILE *fo)
{
	Word i;

	i = 0;			/* Init index */
	do {
		WritePackBits(One3DOLine[i],ColorBits3DO);	/* Save the pixels */
	} while (++i<Width);
	FlushPackBits();		/* Flush the stream */
	i = SkipLineSize-PIndex;	/* Any padding? */
	while (i) {
		PackedLine[PIndex] = 0;	/* Pad it out */
		++PIndex;
		--i;					/* Count down */
	}
	fwrite(PackedLine,1,PIndex,fo);	/* Save it the easy way! */
}

/***************************************

	Remove color zero from the PLUT

***************************************/

static void ZapColor0(void)
{
	Word i;
	i = 0;
	do {
		if (!Pal3DO[i]) {		/* Black? */
			Pal3DO[i] = 0x0001;		/* Force non black */
		}
	} while (++i<256);			/* All colors processed */
}

/***************************************

	Count the number of colors in a shape

***************************************/

static void CountColors(Word8 *ShapePtr,Word8 *MaskPtr,Word Length)
{
	Word i,Count;
	memset(ColorUsed,0,sizeof(ColorUsed));
	memset(Translate,0,sizeof(Translate));
	memset(PLUT,0,sizeof(PLUT));
	if (MaskPtr) {
		while (Length) {
			if (!MaskPtr[0]) {		/* Can I see this? */
				ColorUsed[ShapePtr[0]] = TRUE;	/* Use the color */
			}
			++ShapePtr;
			++MaskPtr;
			--Length;
		}

	} else {
	 	while (Length) {
			ColorUsed[ShapePtr[0]] = TRUE;	/* Just use all pixel data */
			++ShapePtr;
			--Length;
		}
	}
	i = 0;
	Count = 0;
	if (MaskPtr && !PackedFlag) {		/* Reserve color #0 for the invisible color */
		++Count;
	}
	do {
		if (ColorUsed[i]) {		/* Is this color used? */
			Translate[i] = static_cast<Word8>(Count);		/* Init the translate table */
			PLUT[Count] = BigEndian::Load(Pal3DO[i]);		/* Set the plut entry */
			++Count;
		}
	} while (++i<256);		/* All entries scanned? */
	ColorCount = Count;
	switch (ColorBits3DO) {
	case 1:
		i = 2;
		break;
	case 2:
		i = 4;
		break;
	case 4:
		i = 16;
		break;
	case 6:
		i = 32;
		break;
	default:
		i = 256;
	}
	if (ColorCount>i) {
		sprintf((char *)ColorUsed,"Too many colors, found %u, maximum %u",ColorCount,i);
		PrintError((char *)ColorUsed);
	}
}

/***************************************

	Save the PLUT to disk

***************************************/

static void SaveColors(FILE *fo)
{
	Word i;
	i = ColorCount*2;
	if (i&2) {
		i+=2;		/* Long word align! */
	}
	fwrite(PLUT,1,i,fo);
}


/***************************************

	Write out a 3DO shape
	I will also apply compression if needed

***************************************/

static void Write3DOLine(Word Width,FILE *fo)
{
	Word RunCount;
	Word i;
	Word LastPack;
	Word8 *DestPtr;
	Word16 *Ptr3DO;

	PBitCount = 0;		/* Init the packed output routine */
	PIndex = 0;			/* Init the packed output index */

	if (!PackedFlag) {		/* No compression? */
		WriteUnpacked(Width,fo);	/* Just write it out */
		return;
	}

	++PIndex;
	if (ColorBits3DO == 16 || ColorBits3DO==8) {
		++PIndex;
	}

	DestPtr = &PackedLine[PIndex];		/* Init the destination pointer */
	Ptr3DO = &One3DOLine[0];		/* Init the source pointer */
	while (Width) {				/* Continue until done */
		if (Width==1) {			/* Only one cell left? */
			if (Ptr3DO[0]!=0x8000) {	/* Pack as is... */
				WritePackBits(0|0x40,8);		/* Uncompressed pack token */
				WritePackBits(Ptr3DO[0],ColorBits3DO);
				break;
			}
			WritePackBits(0|0x00,2);	/* Save an EOL */
			break;
		}
		RunCount = 0;			/* Init the run length */
		LastPack = Ptr3DO[0];		/* Init the first entry */
		i = 1;
		if (Ptr3DO[1] == LastPack) {	/* Shall I scan for repeater? */
			do {
				++RunCount;			/* I found one */
				++i;
				if (RunCount>=63 || i>=Width) {
					break;
				}
			} while (Ptr3DO[i] == LastPack);	/* Still repeat? */
			if (LastPack!=0x8000) {
				WritePackBits(RunCount|0xC0,8);		/* Repeat pack token */
				WritePackBits(LastPack,ColorBits3DO);	/* Packed value */
			} else {
				if (RunCount+1 == Width) {	/* No more data? */
					WritePackBits(0|0x00,2);
					break;
				}
	  			WritePackBits(RunCount|0x80,8);		/* Invisible pack token */
			}
			++RunCount;			/* Fix to true count */
			Ptr3DO+=RunCount;	/* Adjust source */
			Width-=RunCount;	/* Adjust remainder */
		} else {			/* Scan here for raw data */
			if (LastPack == 0x8000) {
				WritePackBits(0|0x80,8);	/* Single blank char */
				RunCount=1;
			} else {
				for (;;) {
					LastPack = Ptr3DO[i];
					if (LastPack==0x8000) {	/* Do NOT pack in mask shapes! */
						break;
					}
					++RunCount;
					++i;
					if (RunCount>=63 || i>=Width) {
						break;
					}
					if (Ptr3DO[i] == LastPack) {
						if (Ptr3DO[i+1] == LastPack) {
							--RunCount;
							break;
						}
					}
				}
				WritePackBits(RunCount|0x40,8);		/* Uncompressed pack token */
				++RunCount;
				i = 0;
				do {
					WritePackBits(Ptr3DO[i],ColorBits3DO);
				} while (++i<RunCount);
			}
			Ptr3DO+=RunCount;
			Width-=RunCount;
		}
	}

/* Finally!!! The wrap up! */

	if (PBitCount&7) {		/* Odd sized cel? */
		WritePackBits(0,2);	/* Append an EOL */
	}
	FlushPackBits();	   	/* Pad it out */
	if (PIndex<8) {		   	/* Too small? */
		RunCount = 0;	   	/* Just use zero */
		i = 8;
	} else {
		i = (PIndex+3)&(~3);	/* Round to nearest longword */
		RunCount = (i/4)-2;		/* Word count - 2 */
	}
	if (ColorBits3DO==16 || ColorBits3DO==8) {
		PackedLine[0] = static_cast<Word8>(RunCount>>8);	/* Use 16 bits */
		PackedLine[1] = static_cast<Word8>(RunCount);
	} else {
		PackedLine[0] = static_cast<Word8>(RunCount);		/* Use 8 bits */
	}
	while (PIndex<i) {			/* Zero out the rest of the line */
		PackedLine[PIndex] = 0;
		++PIndex;
	}
	fwrite(PackedLine,1,i,fo);	/* Save to disk */
}

/***************************************

	Write out a 16 bit uncoded shape

***************************************/

static void Process16Bit(Word Width,Word Height,Word8 *ShapePtr,Word8 *MaskPtr,
	FILE *fo)
{
	Word x,y;
	Word16 Pixel;

	if (!MaskPtr) {				/* Is it unmasked? */
		y = 0;
		do {
			x = 0;
			do {
				One3DOLine[x] = Pal3DO[ShapePtr[0]];
				++ShapePtr;
			} while (++x<Width);
			Write3DOLine(Width,fo);
		} while (++y<Height);
	} else {
		y = 0;
		do {
			x = 0;
			do {
				if (!MaskPtr[0]) {
					Pixel = Pal3DO[ShapePtr[0]];
				} else {
					Pixel = 0x8000;		/* Invisible color */
				}
				One3DOLine[x]=Pixel;	/* Save the pixel */
				++MaskPtr;
				++ShapePtr;
			} while (++x<Width);
			Write3DOLine(Width,fo);
		} while (++y<Height);
	}
}

/***************************************

	Write out coded shape

***************************************/

static void ProcessOtherBit(Word Width,Word Height,Word8 *ShapePtr,Word8 *MaskPtr,
	FILE *fo)
{
	Word x,y;
	Word16 Pixel;

	if (!MaskPtr) {				/* Is it unmasked? */
		y = 0;
		do {
			x = 0;
			do {
				One3DOLine[x] = Translate[ShapePtr[0]];
				++ShapePtr;
			} while (++x<Width);
			Write3DOLine(Width,fo);
		} while (++y<Height);
	} else {
		y = 0;
		do {
			x = 0;
			do {
				if (!MaskPtr[0]) {
					Pixel = Translate[ShapePtr[0]];
				} else {
					Pixel = 0x8000;		/* Invisible color */
				}
				One3DOLine[x]=Pixel;	/* Save the pixel */
				++MaskPtr;
				++ShapePtr;
			} while (++x<Width);
			Write3DOLine(Width,fo);
		} while (++y<Height);
	}
}

/***************************************

	Convert a shape to 3DO info and save it
	I also can color reduce and apply compression

***************************************/

void Save3DOShape(MyShape_t *MyShapePtr,FILE *fo)
{
	Word Width,Height;
	Word x;
	Word8 *MaskPtr,*ShapePtr;
    Word32 Mark;

	Make3DOPal(MyShapePtr->ShapePal);		/* Create the 3DO Palette (Or PLUT) */
	Width = MyShapePtr->ShapeWidth;			/* Get the size of the shape */
	Height = MyShapePtr->ShapeHeight;
	ShapePtr = MyShapePtr->ShapePtr;	/* Save the data */
	MaskPtr = MyShapePtr->MaskPtr;
	if (MaskPtr) {
		ZapColor0();			/* Adjust the PLUT to remove color 0,0,0 */
	}
	if (ColorBits3DO!=16) {
		CountColors(ShapePtr,MaskPtr,Width*Height);		/* Count the colors and fix the translate tables */
	}

	if (HeaderFlag) {				/* Will I save the CCB? */
		MyCCB.ccb_Flags = 0x47664010;	/* 0x20 = No Mask, 0x200 = Packed */
		if (ColorBits3DO==16) {
			MyCCB.ccb_PLUTPtr = 0;
			MyCCB.ccb_SourcePtr = 0x30;
			MyCCB.ccb_Flags &= ~0x800000;		/* Don't load plut */
		} else {
			MyCCB.ccb_PLUTPtr = 0x2C;	/* Index to data after CCB */
			Mark = ColorCount*2;
			if (Mark&2) {
				Mark+=2;
			}
			MyCCB.ccb_SourcePtr = (Word32)(Mark+0x30);
			MyCCB.ccb_Flags |= 0x800000;		/* Load the plut */
		}
		MyCCB.ccb_PRE0 = (Height-1)<<6;	/* Save the height of the shape */
		if (!MyShapePtr->MaskPtr) {		/* Is there a mask? */
			MyCCB.ccb_Flags |= 0x20;	/* Disable the masking */
			MyCCB.ccb_PRE0 |= 0x40000000;	/* Allow use of 000 */
		} else {
			MyCCB.ccb_Flags &= ~0x20;	/* Enable the mask */
		}

			/* Create preable #0 */

		if (ColorBits3DO==16) {
			MyCCB.ccb_PRE0 |= 0x10;			/* This CCB is not coded */
		}
		switch (ColorBits3DO) {
		default:
			x=6;		/* 16 bits per pixel */
			break;
		case 1:			/* 1 bit per pixel */
			x=1;
			break;
		case 2:			/* 2 bits per pixel */
			x=2;
			break;
		case 4:			/* 4 bits per pixel */
			x=3;
			break;
		case 6:			/* 6 bits per pixel */
			x=4;
			break;
		case 8:			/* 8 bits per pixel */
			x=5;
		}
		MyCCB.ccb_PRE0 |= x;		/* Set the bits per pixel */

		Mark = Width*ColorBits3DO;	/* Number of bits per line of data */
		Mark += 0x1F;				/* Round to the nearest long word */
		Mark &= ~0x1F;
		SkipLineSize = (Word)(Mark/8UL);		/* Number of bytes per line */
		if (SkipLineSize<8) {
			Mark=8*8;				/* Minimum 8 bytes per line */
			SkipLineSize=8;			/* Minimum of 8 bytes per line */
		}
		Mark/=32;					/* Number of longwords per line */
		Mark-=2;					/* Remove 2 from the count */
		Mark<<=16;					/* Move to place in CCB */
		if (ColorBits3DO<8) {
			Mark<<=8;			/* For 1,2,4 and 6 bit CCBs */
		}
		MyCCB.ccb_PRE1 = Mark;		/* Save the preamble */
		MyCCB.ccb_PRE1 |= ((Width-1)&0x7FF);	/* Get pixel width */
		MyCCB.ccb_PRE1 |= 0x1000;		/* Leave the blue pixel alone */

		if (PackedFlag) {				/* Allow packing */
			MyCCB.ccb_Flags |= (1<<9);	/* Set the packed flag */
		} else {
			MyCCB.ccb_Flags &= ~(1<<9);
		}

/* The CCB header is BIG ENDIAN, make sure it's saved correctly */

#if !defined(BURGER_BIGENDIAN)		/* Save off the CCB header */
		{
		Word32 *LongPtr;	/* Swap the ending since this tool is on */
							/* a little endian machine */
		LongPtr = (Word32 *)&MyCCB;
        x = 0;
        do {
        	Mark = SwapEndian::Load(LongPtr[0]);	/* Swap endian */
			fwrite(&Mark,1,4,fo);	/* Save the longword */
			++LongPtr;
		} while (++x<15);		/* All done? */
		}
#else
		fwrite(&MyCCB,1,sizeof(MyCCB),fo);	/* Save the header as is */
#endif
	}


	/* Prepare to save the data */

	if (ColorBits3DO==16) {			/* 16 bit data */
		Process16Bit(Width,Height,ShapePtr,MaskPtr,fo);
	} else {
		SaveColors(fo);
		ProcessOtherBit(Width,Height,ShapePtr,MaskPtr,fo);
	}


/* As the final coup de grace, I must pad the data to make SURE */
/* it is long word aligned or I will get a page fault */

	Mark = ftell(fo)&3;	/* I need to long word align the file */
	if (Mark) {			/* Padding needed? */
		do {
			fputc(0,fo);	/* Pad it... */
		} while (++Mark<4);	/* Finish the job! */
	}
}


