#include "Slicer.h"
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

/***************************************

	Tables for color lookup

***************************************/

static Word ColorMapTbl[256];		/* Colors found */
static Word FirstColor;				/* First available color */

/***************************************

	Remap the colors found into a color remap table

***************************************/

static Word8 Remap(Word Input)
{
	Word New;

	Input = TransTable[Input];		/* Translate by user table */
	New = ColorMapTbl[Input];		/* Is the color already found */
	if (New!=(Word)-1) {
		return New&0x0F;		/* Return the 16 color entry */
	}
	New = FirstColor;		/* Assign a new color */
	ColorMapTbl[Input] = New;	/* Save in the table */
	++FirstColor;		/* Add a color */
	if (New==16) {		/* Too many assigned!! */
		PrintError("Too Many colors!!!");
	}
	return New&0xF;		/* Return the new color */
}

/***************************************

	Translate a single line from 256 color to 16 color

***************************************/

static Word8 *Copy1Line(Word8 *WorkPtr,Word8 *ArtPtr,Word Width)
{
	Word Temp;
	do {
		Temp = Remap(ArtPtr[0]);	/* Remap the pixel */
		WorkPtr[0] = static_cast<Word8>(Temp<<4);		/* Save the upper pixel */
		if (!--Width) {			/* Odd pixel width? */
			++WorkPtr;			/* Pad the pointer */
			break;
		}
		Temp = Remap(ArtPtr[1]);	/* Remap the next pixel */
		WorkPtr[0] |= Temp;		/* Blend */
		++WorkPtr;			/* Accept the pixel */
		ArtPtr+=2;			/* Adjust the source pointer */
	} while (--Width);		/* More? */
	return WorkPtr;			/* Return the dest pointer */
}

/***************************************

	Save the RGB colors

***************************************/

static Word8 *SaveRGB(Word8 *WorkPtr)
{
	Word Count;
	Word Color;

	Count = FirstColor;
	if (Count>=17) {
		Count=16;
	}
	WorkPtr[0] = (Word8)Count;
	++WorkPtr;
	Color = 0;
	do {
		Word i;
		i = 0;
		do {
			if (ColorMapTbl[i]==Color) {
				i=i+i+i;
				WorkPtr[0] = g_pPalette[i];
				WorkPtr[1] = g_pPalette[i+1];
				WorkPtr[2] = g_pPalette[i+2];
				WorkPtr+=3;
				break;
			}
		} while (++i<256);
		++Color;
	} while (--Count);
	return WorkPtr;
}

/***************************************

	Cut out a tileset, the tileset is only in the form of...

	SLICEFONT FontX,FontY,CelHeight,FirstCelASCII,TotalCels,
		FileName

	There is no header saved, it will create a monospaced tile font
	with no header or palette information. This is for simple Ultima type
	fonts for limited use.

***************************************/

void SliceFont(void)
{
	Word ShapeBox[5];	/* Rect for the current shape */
	Word8 *ShapePtr;		/* Pointer to new shape */
	Word8 *WidthPtr;		/* Pointer to the font width table */
	Word8 *ArrayPtr;		/* Pointer to the strike offset table */
	Word8 *WidthPixels;	/* Pointer to the shape width pixel data */
	Word8 *WorkPtr;		/* End of memory pointer */
	char *TextPtr;		/* Pointer to the dest filename */
	Word i;             /* Temp */
	Word Index;			/* Current font number being cut */
	Word BaseX;			/* X coord of font being cut */

	if (!g_pPictureBuffer) {			/* Art loaded? */
		PrintError("No source art loaded");
		return;
	}

	memset(ColorMapTbl,-1,sizeof(ColorMapTbl));		/* Init the table */
	ColorMapTbl[0] = 0;			/* Map zero to ZERO */
	FirstColor = 1;				/* First remapped color */

	i = 0;			/* Init index to ShapeBox */
	do {
		TextPtr = strtok(0,NumDelimiters);
		if (!TextPtr) {
			PrintError("Not enough parms for SLICEFONT");
			return;
		}
		if (!isdigit(TextPtr[0])) {		/* First char must be a digit */
			PrintError("Non-numeric parm for SLICEFONT");
			return;
		}
		ShapeBox[i] = AsciiToInteger(TextPtr);		/* Get the number */
	} while (++i<5);			/* All 6 parms fetched? */
	TextPtr = strtok(0,NumDelimiters);	/* Get the filename */

	if (!ShapeBox[2]) {
		PrintError("Can't cut a zero height font");
		return;
	}
	if (!ShapeBox[4]) {
		PrintError("No cels in the font");
		return;
	}
	ShapePtr = (Word8 *)Alloc((PicWidth*PicHeight)+768);   /* Get the memory */

	if (!ShapePtr) {
		PrintError("Not enough memory for SLICEFONT");
		return;
	}

	WorkPtr = ShapePtr;		/* Init my work pointer */
	((Word16 *)WorkPtr)[0] = LittleEndian::Load(static_cast<Word16>(ShapeBox[2]));	/* Height of font */
	((Word16 *)WorkPtr)[1] = LittleEndian::Load(static_cast<Word16>(ShapeBox[4]));	/* Number of cels */
	((Word16 *)WorkPtr)[2] = LittleEndian::Load(static_cast<Word16>(ShapeBox[3]));	/* ASCII Start */

	WorkPtr+=6;		/* Index past header */

	WidthPtr = WorkPtr;		/* Set pointer to width table */
	WorkPtr += ShapeBox[4];	/* Index past the cel index */
	ArrayPtr = WorkPtr;		/* Set pointer to offset table */
	WorkPtr += (ShapeBox[4]+1)*2;	/* Index past the array */

	BaseX = ShapeBox[0];	/* Init the X coord */
	WidthPixels = &g_pPictureBuffer[((ShapeBox[1]+ShapeBox[2])*PicWidth)];
	Index = 0;

	do {
		Word Width;         /* Width of cut */
		Word8 *ArtPtr;		/* Pointer to the source art */
		Word Height;		/* Height of the shape to cut */

		ArtPtr = &g_pPictureBuffer[(ShapeBox[1]*PicWidth)+BaseX];
		Height = ShapeBox[2];       /* Store my counters */

		Width = 0;
		do {
			if (BaseX>=PicWidth) {		/* Off the screen? */
				if (Width) {		/* Already have a cell? */
					break;			/* Accept it */
				}
				PrintError("Pixel width table is too small");
				Free(ShapePtr);
				return;		/* Error! */
			}
			++BaseX;		/* Move one pixel */
			++Width;		/* Increase the width */
		} while (WidthPixels[BaseX]!=1);
		WidthPtr[Index] = static_cast<Word8>(Width);
		i = static_cast<Word>(WorkPtr-ArrayPtr);
		((Word16 *)ArrayPtr)[Index] = LittleEndian::Load(static_cast<Word16>(i));

		do {
			WorkPtr = Copy1Line(WorkPtr,ArtPtr,Width);	/* Copy a line */
			ArtPtr += PicWidth;		/* Next line down */
		} while (--Height);		/* All done for height */
	} while (++Index<ShapeBox[4]);	/* Next cel */

	i = static_cast<Word>(WorkPtr-ArrayPtr);
	((Word16 *)ArrayPtr)[Index] = LittleEndian::Load(static_cast<Word16>(i));
	WorkPtr = SaveRGB(WorkPtr);
	FileManager::SaveFile(TextPtr,ShapePtr,WorkPtr-ShapePtr);	/* Save the output */
	Free(ShapePtr);		/* Release the buffer */
}

