#include "Slicer.h"
#include <string.h>
#include <stdlib.h>

typedef struct {
	Word16 width;
	Word16 height;
	Word16 leftoffset;
	Word16 topoffset;
	Word16 columnofs[1];
} Sprite2D_t;

typedef struct {
	Word8 Start;
	Word8 Count;
	Word16 Offset;
} Run_t;

static Word8 OneLine[1024];	/* One line of data */
static Word RunSize[1024];	/* Size in bytes of each run */
static Run_t *RunPtrs[1024];	/* Pointer to each run */
static Word8 *PixelData;		/* Pointer to pixel data */
static Word8 *PixelDataPtr;	/* Running pointer to pixel data */
static Word8 *RunData;		/* Pointer to Run_t structures */
static Word8 *RunDataPtr;	/* Running pointer to Run_t structures */

/***************************************

	Convert all the pixel offsets from
	PixelDataPtr-PixelData to offsets from the beginning
	of the data.

***************************************/

static void AdjustRun(Run_t *RunPtr,Word AdjustValue)
{
	do {
		RunPtr->Offset += static_cast<Word16>(AdjustValue);		/* Adjust the pointer */
		++RunPtr;			/* Next run */
	} while (RunPtr->Start!=255);	/* End of the data? */
}

/***************************************

	Swap the endian of a pixel run

***************************************/

static void SwapEndianRun(Run_t *RunPtr)
{
	do {
		RunPtr->Offset = SwapEndian::Load(RunPtr->Offset);		/* Adjust the pointer */
		++RunPtr;			/* Next run */
	} while (RunPtr->Start!=255);	/* End of the data? */
}

/***************************************

	I will now add a run to the run list

***************************************/

static void AddRun(Word Start,Word Count)
{
	Run_t *NewRunPtr;

	NewRunPtr = (Run_t *)RunDataPtr;	/* Get free memory */
	RunDataPtr = (Word8 *)(NewRunPtr+1);	/* Store adjusted pointer */
	NewRunPtr->Start = static_cast<Word8>(Start);	/* Top Y coord */
	NewRunPtr->Count = static_cast<Word8>(Count);	/* Number of pixels in run */
	NewRunPtr->Offset = static_cast<Word16>(PixelDataPtr-PixelData);	/* Offset to data */
	memcpy(PixelDataPtr,OneLine,Count);	/* Copy data */
	PixelDataPtr+=Count;		/* Adjust pointer */
}

/***************************************

	Create all the vertical strips

***************************************/

static void CreateStrips(Word Width,Word Height,Word8 *ShapePtr,Word8 *MaskPtr)
{
	Word RunCount;
	Word EndShape;
	Word Index;
	Run_t *DataPtr;

	Index = 0;
	EndShape = Width*Height;
	do {
		Word Height2;
		Word TopY;

		Height2 = Index;		/* Init x coord */
		RunCount = 0;
		DataPtr = (Run_t *)RunDataPtr;
		TopY = 0;
		do {
			if (!MaskPtr[Height2]) {		/* Start of a run? */
				Word8 *StartPtr;

				StartPtr = OneLine;
				do {
					if (MaskPtr[Height2]) {	/* End the run now? */
						break;
					}
					StartPtr[0] = ShapePtr[Height2];
					++StartPtr;
					Height2+=Width;
				} while (Height2<EndShape);
				AddRun(TopY,static_cast<Word>(StartPtr-OneLine));
				++RunCount;
				TopY = TopY + static_cast<Word>(StartPtr-OneLine);
			}
			Height2+=Width;
			++TopY;
		} while (Height2<EndShape);
		if (RunCount) {
			RunDataPtr[0] = 255;	/* Mark as end record */
			++RunDataPtr;
			RunSize[Index] = (RunCount*4)+1;
			RunPtrs[Index] = DataPtr;
		} else {
			RunSize[Index] = 0;
		}
	} while (++Index<Width);
}

/***************************************

	Convert a shape to 3DO info and save it
	I also can color reduce and apply compression

***************************************/

void SaveScaleShape(MyShape_t *MyShapePtr,FILE *fo,Word Little)
{
	Word Width,Height;
	Word8 *MaskPtr,*ShapePtr;
	Word i,j;

	Width = MyShapePtr->ShapeWidth;			/* Get the size of the shape */
	Height = MyShapePtr->ShapeHeight;
	ShapePtr = MyShapePtr->ShapePtr;	/* Save the data */
	MaskPtr = MyShapePtr->MaskPtr;

	RunData = (Word8 *)Alloc(sizeof(Run_t)*10000);
	RunDataPtr = RunData;
	PixelData = (Word8 *)Alloc(Width*Height);
	PixelDataPtr = PixelData;

	if (Little) {
		WriteLittle(fo,(Word16)Width);		/* Save the shape's width */
		WriteLittle(fo,(Word16)Height);
		WriteLittle(fo,(Word16)(((MyShapePtr->OriginalShapeWidth+MyShapePtr->AdjustX)/2)-MyShapePtr->ReduceX));
		WriteLittle(fo,(Word16)((MyShapePtr->OriginalShapeHeight+MyShapePtr->AdjustY)-MyShapePtr->ReduceY));
	} else {
		WriteBig(fo,(Word16)Width);		/* Save the shape's width */
		WriteBig(fo,(Word16)Height);
		WriteBig(fo,(Word16)(((MyShapePtr->OriginalShapeWidth+MyShapePtr->AdjustX)/2)-MyShapePtr->ReduceX));
		WriteBig(fo,(Word16)((MyShapePtr->OriginalShapeHeight+MyShapePtr->AdjustY)-MyShapePtr->ReduceY));
	}

	CreateStrips(Width,Height,ShapePtr,MaskPtr);		/* Create all the strips */

	/* Now that the shape is sliced, let's save it off */
	/* First comes the offset table */

	i = 0;
	j = 8+(Width*2);		/* Init the index (Header+Table) */
	do {
		if (RunSize[i]) {
			if (Little) {
				WriteLittle(fo,(Word16)j);	/* Save off the index */
			} else {
				WriteBig(fo,(Word16)j);
			}
		} else {
			WriteBig(fo,(Word16)0);		/* No index at all... */
		}
		j = j+RunSize[i];	/* Index to the next run */
	} while (++i<Width);		/* All done? */

	/* Now adjust the strips to point to the actual data in the file */

	i = 0;
	do {
		if (RunSize[i]) {	/* Any data? */
			AdjustRun(RunPtrs[i],j);	/* Adjust the run offsets */
#if defined(BURGER_BIGENDIAN)
			if (Little) {
#else
			if (!Little) {
#endif
				SwapEndianRun(RunPtrs[i]);
			}
			fwrite(RunPtrs[i],1,RunSize[i],fo);	/* Save the runs */
		}
	} while (++i<Width);

	fwrite(PixelData,1,PixelDataPtr-PixelData,fo);	/* Save all the run data */
	Free(RunData);		/* Release the data */
	Free(PixelData);
}
