/***************************************

	Slicer by Rebecca Ann Heineman (Mac/IIgs/3DO/IBM)

	This will extract art from any source

***************************************/

#include "Slicer.h"
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

/***************************************

	Global data for me

***************************************/

char InputLine[256];        /* Input line from script */
Bool InitPalette = TRUE;

/***************************************

	Target machine globals

***************************************/

const char *MachNames[] = {
	"IBM","MAC","IIGS","SNES","3DO","DOOM3DO","WIN95",
	"SCALEIBM","SCALEMAC"
};
typedef enum {
	PCIBM,PCMAC,PCIIGS,PCSNES,PC3DO,PC3DODOOM,PCWIN95,
	PCSCALEIBM,PCSCALEMAC,PCNONE
	} Machines;     /* Type of machine to save to */
Machines SaveType = PCIBM;  /* Assume IBM */

/***************************************

	Input art types

***************************************/

const char *ArtNames[] = {
	"LBM","TGA","PNT","BMP","RLE","TGC","GIF","PCX"
};
typedef enum {
	ARTLBM,ARTTGA,ARTPNT,ARTBMP,ARTRLE,ARTTGC,ARTGIF,ARTPCX,ARTNONE
	} ArtFiles;     /* Type of machine to save to */

#define CommandCount 15      /* Number of commands */
const char *Commands[] = {
	"MACHINE","ARTTYPE","LOAD","PALETTE","SLICE","SAVE",
	"MASK","TRANSPOSE","REDUCE","SHAPEHEADER","SIDEWAYS",
	"SLICETILE","PACKED","INDEXARRAY","SLICEFONT"
};

/***************************************

	What machine shall I save for?

***************************************/

static void LogMachine(void)
{
	char *TextPtr;      /* Pointer to token */
	Word i;     /* Index */

	TextPtr = strtok(0,Delimiters);     /* Get the operand */
	if (!TextPtr) {
		PrintError("No machine name in MACHINE");   /* No operand! */
		return;
	}
	i = 0;          /* First index */
	do {
		if (!StringCaseCompare(TextPtr,MachNames[i])) {    /* Match? */
			SaveType = (Machines)i;               /* Save the machine type */
			return;
		}
	} while (++i<PCNONE);               /* No machine? */
	PrintError("Not supported machine");    /* Not supported! */
}

/***************************************

	Use masks?

***************************************/

static void LogMask(void)
{
	char *TextPtr;

	TextPtr = strtok(0,Delimiters);     /* Get the operand */
	if (!TextPtr) {
		PrintError("No operand for MASK");  /* No operand! */
		return;
	}
	if (!StringCaseCompare(TextPtr,"OFF")) {   /* No masking allowed! */
		MaskEnable = FALSE;
		return;
	}
	MaskValue = AsciiToInteger(TextPtr);      /* Convert to value */
	MaskEnable = TRUE;
}

/***************************************

	Use transposition?

***************************************/

static void LogTranspose(void)
{
	char *TextPtr;
	Word Index,NewVal;

	TextPtr = strtok(0,NumDelimiters);      /* Get the operand */
	if (!TextPtr) {
		PrintError("No operand for TRANSPOSE"); /* No operand! */
		return;
	}
	if (!StringCaseCompare(TextPtr,"OFF")) {   /* No masking allowed! */
		InitTranspose();
		return;
	}
	Index = AsciiToInteger(TextPtr);      /* Convert to value */

	TextPtr = strtok(0,NumDelimiters);
	if (!TextPtr) {
		PrintError("Not enough parms for TRANSPOSE");
		return;
	}
	NewVal = AsciiToInteger(TextPtr);
	TransTable[Index] = (Word8)NewVal;
}

/***************************************

	Use reduction?

***************************************/

static void LogReduce(void)
{
	char *TextPtr;

	TextPtr = strtok(0,NumDelimiters);      /* Get the operand */
	if (!TextPtr) {
		PrintError("No operand for REDUCE");    /* No operand! */
		return;
	}
	if (!StringCaseCompare(TextPtr,"OFF")) {   /* No masking allowed! */
		ReduceEnable = 0;
		return;
	}
	ReduceEnable = 1;
	ReduceColor = AsciiToInteger(TextPtr);        /* Convert to value */
}

/***************************************

	Enable / disable shape headers

***************************************/

static void LogHeader(void)
{
	OffOn("SHAPEHEADER",&HeaderFlag);
}

/***************************************

	Enable / disable auto index array

***************************************/

static void LogIndexArray(void)
{
	OffOn("INDEXARRAY",&IndexArrayFlag);
}


/***************************************

	Enable / disable game machine packed data

***************************************/

static void LogPacked(void)
{
	char *TextPtr;
	Word TempColors;

	OffOn("PACKED",&PackedFlag);
	if (SaveType!=PC3DO && SaveType!=PC3DODOOM) {	/* The 3DO has special features */
		return;
	}
	TextPtr = strtok(0,NumDelimiters);
	if (!TextPtr) {
		ColorBits3DO = 16;		/* If no delimiter, then assume 16 bit color */
	}
	TempColors = AsciiToInteger(TextPtr);	/* Get the number of colors to parse */
	if (TempColors>=33) {
		ColorBits3DO = 16;
	} else if (TempColors>=17) {
		ColorBits3DO = 6;
	} else if (TempColors>=5) {
		ColorBits3DO = 4;
	} else if (TempColors>=3) {
		ColorBits3DO = 2;
	} else {
		ColorBits3DO = 1;
	}
}

/***************************************

	Enable / disable 90 degree rotation

***************************************/

static void LogSideways(void)
{
	OffOn("SIDEWAYS",&SidewaysFlag);
}

/***************************************

	Load in the art file

***************************************/

static void LoadArt(void)
{
	RGBAWord8_t TempPalette[256];

	if (InitPalette) {		/* Has the palettes been initialized? */
		InitPalette = FALSE;
		g_pPalette = Palettes;
	} else {
		g_pPalette += 768;		/* Index to the next palette buffer */
	}
	char *FileName = strtok(0,Delimiters);        /* Get the operand */
	if (!FileName) {
		PrintError("No filename in LOAD");  /* No operand! */
		return;
	}
	Word ArtType = 0;
	WordPtr uNameLength = StringLength(FileName);
	if (uNameLength<3) {
		PrintError("Can't load data format, not supported.");
		return;
	}
	uNameLength -= 3;
	do {
		if (!StringCaseCompare(&FileName[uNameLength],ArtNames[ArtType])) {
			break;
		}
	} while (++ArtType<ARTNONE);
	if (ArtType==ARTNONE) {
		PrintError("Can't load data format, not supported.");
		return;
	}
	InputMemoryStream Data;
	if (Data.Open(FileName)) {	// Load in the art file
		PrintError("Can't load art file");  /* Not supported! */
		return;
	}
	if (g_pPictureBuffer) {		/* Was an art file already loaded? */
		Free(g_pPictureBuffer);	/* Release it! */
		g_pPictureBuffer = NULL;	/* Zap the buffer ptr */
	}
	Image *pImage = NULL;
	Word RetVal=2;
	switch (ArtType) {      /* Load in with the proper art loader */
	case ARTLBM:
		{
			FileLBM ParserLBM;
			pImage = ParserLBM.Load(&Data);		// Read in an LBM file
			MemoryCopy(TempPalette,ParserLBM.GetPalette(),sizeof(TempPalette));
		}  
		break;
	case ARTTGA:
	case ARTTGC:
		{
			FileTGA ParserTGA;
			pImage = ParserTGA.Load(&Data);		// Read in an TGA file
			MemoryCopy(TempPalette,ParserTGA.GetPalette(),sizeof(TempPalette));
		}
		break;
	case ARTPCX:
		{
			FilePCX ParserPCX;
			pImage = ParserPCX.Load(&Data);		// Read in an PCX file
			MemoryCopy(TempPalette,ParserPCX.GetPalette(),sizeof(TempPalette));
		}
		break;
	case ARTGIF:
		{
			FileGIF *pParserGIF = new (Alloc(sizeof(FileGIF))) FileGIF;
			pImage = pParserGIF->Load(&Data);		// Read in an GIF file
			MemoryCopy(TempPalette,pParserGIF->GetPalette(),sizeof(TempPalette));
			Delete(pParserGIF);
		}
		break;
	case ARTBMP:
	case ARTRLE:
		{
			FileBMP ParserBMP;
			pImage = ParserBMP.Load(&Data);		// Read in an BMP file
			MemoryCopy(TempPalette,ParserBMP.GetPalette(),sizeof(TempPalette));
		}
		break;
	default :
		RetVal = 2;         /* Error! */
	}

	if (pImage) {		/* No error? */
		g_pPictureBuffer = pImage->GetImage();
		Word uColor = 256;
		const RGBAWord8_t *pColor = TempPalette;
		Word8 *pPalette = g_pPalette;
		do {
			pPalette[0] = pColor->m_uRed;
			pPalette[1] = pColor->m_uGreen;
			pPalette[2] = pColor->m_uBlue;
			++pColor;
			pPalette += 3;
		} while (--uColor);
		PicWidth = pImage->GetWidth();
		PicHeight = pImage->GetHeight();
		Free(pImage);			// Deliberately delete the class but not delete the allocated image
		RetVal = 0;
	}
	if (RetVal!=2 && !pImage) {		/* Normal error */
		PrintError(Globals::GetErrorMsg());	/* Print the error */
	}
	if (RetVal==2) {	/* Bad file type */
		PrintError("Can't load data format, not supported.");
	}
}

/***************************************

	Save the palette file

***************************************/

static void SavePalette(void)
{
	Word Flag;
	char *FileName;
	Word i,j;
	Word RGBColor;

	if (!g_pPictureBuffer) {
		PrintError("No loaded picture");
		return;
	}

	FileName = strtok(0,Delimiters);        /* Get the operand */
	if (!FileName) {
		PrintError("No filename in SAVEPAL");   /* No operand! */
		return;
	}
	Flag = TRUE;		/* Assume I didn't save */
	Globals::SetErrorMsg("");		/* Init message */
	switch (SaveType) {
	case PC3DO :
	case PC3DODOOM :
		Flag = FileManager::SaveFile(FileName,Default3DOPal,128);
		break;
	case PCSCALEIBM :
	case PCSCALEMAC :
	case PCMAC :
	case PCWIN95 :
	case PCIBM :
		Flag = FileManager::SaveFile(FileName,g_pPalette,768);      /* Mac uses it as is */
		break;
	case PCIIGS : {
		Word16 IIGSPal[16];
		i = 0;          /* Init counter */
		j = 0;
		do {
			RGBColor = (g_pPalette[i]<<4)&0xF00;       /* Convert to IIGS RGB */
			RGBColor |= g_pPalette[i+1]&0xF0;
			RGBColor |= g_pPalette[i+2]>>4;
			IIGSPal[j] = LittleEndian::Load((Word16)RGBColor);
			i+=3;
		} while (++j<16);
		Flag = FileManager::SaveFile(FileName,IIGSPal,32);
		}
		break;
	case PCSNES : {
		Word16 SNESPal[16];
  		i = 0;          /* Init counter */
  		j = 0;
  		do {
  			RGBColor = (g_pPalette[i]<<7)&0x7C00;      /* Convert to SNES RGB */
  			RGBColor |= (g_pPalette[i+1]<<2)&0x3E0;
  			RGBColor |= g_pPalette[i+2]>>3;
  			SNESPal[j] = LittleEndian::Load((Word16)RGBColor);
			i+=3;
		} while (++j<16);
		Flag = FileManager::SaveFile(FileName,SNESPal,32);
		}
		break;
	}
	if (Flag) {			/* Error? */
		if (StringLength(Globals::GetErrorMsg())) {		/* Failsafe! */
			PrintError(Globals::GetErrorMsg());
		}
	}
}

/***************************************

	Scan the edges of a shape and reduce it
	Sets ReduceX,ReduceY

***************************************/

static void ShrinkShape(MyShape_t *MyShapePtr)
{
	Word8 *ShapePtr,*MaskPtr;
	Word8 *MyPtr,*MyPtr2;
	Word Width,Height;
	Word Top,Bottom,Left=0,Right;
	Word NewWidth,NewHeight;
	Word Index;

	MyShapePtr->ReduceX = 0;            /* Assume no change is needed */
	MyShapePtr->ReduceY = 0;

	ShapePtr = MyShapePtr->ShapePtr;       /* Save the pointer to the shape */
	MaskPtr = MyShapePtr->MaskPtr;
	Width = MyShapePtr->ShapeWidth;        /* Init my internal height and width */
	Height = MyShapePtr->ShapeHeight;

	if (!Width || !Height) {    /* Failsafe, can't help zero size shape */
		return;
	}

/* Crop down the top */

	Top = 0;            /* Init top Y coord */
	MyPtr = ShapePtr;
	do {
		Index = Width;  /* Reset the count */
		do {
			if (MyPtr[0]!=ReduceColor) {    /* Top line to clip... */
				goto GOTTOP;    /* I have a hit */
			}
			++MyPtr;    /* Next byte */
		} while (--Index);  /* Width traversed */
	} while (++Top<Height); /* At the bottom? */
	Right = 0;          /* Blank out the shape */
	Top = 0;
	NewWidth = 0;
	NewHeight = 0;
	goto EXIT;

GOTTOP:         /* Top is set to the proper top Y coord */
				/* It is safe to assume that I cannot find an "Empty" shape */
	Bottom = Height-1;
	if (Bottom!=Top) {  /* One line tall? */
		do {
			MyPtr=&ShapePtr[(Word32) Bottom*Width];  /* Init the shape pointer */
			Index = Width;      /* Init width index */
			do {
				if (MyPtr[0]!=ReduceColor) {    /* Bottom line to clip... */
					goto GOTBOTTOM; /* I have a hit */
				}
				++MyPtr;    /* Next byte */
			} while (--Index);
		} while (--Bottom);
	}
GOTBOTTOM:
	++Bottom;           /* Correct bottom line # */
	NewHeight = Bottom-Top;     /* I now have the height */

	Left = 0;           /* Init left side */
	for (;;) {
		MyPtr=&ShapePtr[((Word32)Top*Width)+Left];   /* Init the shape pointer */
		Index = NewHeight;      /* Init width index */
		do {
			if (MyPtr[0]!=ReduceColor) {    /* Bottom line to clip... */
				goto GOTLEFT;   /* I have a hit */
			}
			MyPtr+=Width;   /* Next byte DOWN */
		} while (--Index);
		++Left;
	}
GOTLEFT:        /* I now have the left edge! */
				/* Let's finish the job! */
	Right = Width-1;
	if (Left!=Right) {
		do {
			MyPtr=&ShapePtr[((Word32)Top*Width)+Right];  /* Init the shape pointer */
			Index = NewHeight;      /* Init width index */
			do {
				if (MyPtr[0]!=ReduceColor) {    /* Bottom line to clip... */
					goto GOTRIGHT;  /* I have a hit */
				}
				MyPtr+=Width;   /* Next byte */
			} while (--Index);
		} while (--Right);
	}
GOTRIGHT:
	++Right;            /* Correct right side */
	NewWidth = Right-Left;  /* What's the width of the new shape */

	Height = 0;
	do {
		Index = 0;
		MyPtr = &ShapePtr[(Word32)Height*NewWidth];  /* Dest pointer */
		MyPtr2 = &ShapePtr[(Top+Height)*(Word32)Width+Left]; /* Source pointer */
		do {
			MyPtr[Index] = MyPtr2[Index];   /* Copy a line of data */
		} while (++Index<NewWidth);         /* Line done? */
	} while (++Height<NewHeight);

	if (MaskEnable) {       /* If there is a mask, reduce it too */
		Height = 0;
		do {
			Index = 0;
			MyPtr = &MaskPtr[(Word32)Height*NewWidth];   /* Dest pointer */
			MyPtr2 = &MaskPtr[(Top+Height)*(Word32)Width+Left];  /* Source pointer */
			do {
				MyPtr[Index] = MyPtr2[Index];   /* Copy a line of data */
			} while (++Index<NewWidth);         /* Line done? */
		} while (++Height<NewHeight);
	}
EXIT:
	MyShapePtr->ShapeWidth = NewWidth;
	MyShapePtr->ShapeHeight = NewHeight;
	MyShapePtr->ReduceX = Left;
	MyShapePtr->ReduceY = Top;
}

/***************************************

	Rotate a shape 90 degrees

***************************************/

static void RotateShape(MyShape_t *MyShapePtr)
{
	Word Width,Height;
	Word Width2,Height2;
	Word32 Size;
	Word8 *ShapePtr;
	Word8 *DestPtr;
	Word8 *TempPtr1,*TempPtr2;

	Width = MyShapePtr->ShapeWidth;
	Height = MyShapePtr->ShapeHeight;
	MyShapePtr->ShapeWidth = Height;	/* Save the swapped width and height */
	MyShapePtr->ShapeHeight = Width;
	ShapePtr = MyShapePtr->ShapePtr;
	Size = (Word32)Width * (Word32)Height;
	DestPtr = static_cast<Word8 *>(Alloc(Size));

	TempPtr1 = ShapePtr;		/* Init the source pointer */
	Height2 = 0;				/* Init the height count (For source) */
	do {
		TempPtr2 = &DestPtr[Height2];	/* Init the dest pointer */
		Width2 = 0;				/* Init the width count (For source) */
		do {
			TempPtr2[0] = TempPtr1[Width2];	/* Copy a byte */
			TempPtr2+=Height;	/* Next line down (Sideways, use height not width) */
		} while (++Width2<Width);
		TempPtr1+=Width;		/* Adjust source pointer */
	} while (++Height2<Height);
	memcpy(ShapePtr,DestPtr,Size);		/* Copy the finished shape */

	ShapePtr = MyShapePtr->MaskPtr;	/* Do I have a mask? */

	if (ShapePtr) {			/* Do I have a mask? */
		TempPtr1 = ShapePtr;
		Height2 = 0;				/* Init the height count (For source) */
		do {
			TempPtr2 = &DestPtr[Height2];	/* Init the dest pointer */
			Width2 = 0;				/* Init the width count (For source) */
			do {
				TempPtr2[0] = TempPtr1[Width2];	/* Copy a byte */
				TempPtr2+=Height;	/* Next line down (Sideways, use height not width) */
			} while (++Width2<Width);
			TempPtr1+=Width;		/* Adjust source pointer */
		} while (++Height2<Height);
		memcpy(ShapePtr,DestPtr,Size);
	}
	Free(DestPtr);						/* Release the temp memory */
}

/***************************************

	Cut out the shapes

***************************************/

static void SliceArt(void)
{
	MyShape_t *MyShapePtr;
	Word ShapeBox[4];   /* Rect for the current shape */
	Word8 *ShapePtr;    /* Pointer to new shape */
	Word8 *MaskPtr;     /* Pointer to the mask */
	Word8 *ArtPtr;      /* Pointer to source art */
	char *TextPtr;              /* Pointer to the slice index */
	Word i;             /* Temp */
	Word Height;        /* Height of shape */
	Word Width;         /* Width of cut */
	int AdjustX,AdjustY;
	Word32 ShapeSize;    /* Bytes to cut */
	Word PicByte;       /* Byte of data */

	if (!g_pPictureBuffer) {			/* Art loaded? */
		PrintError("No source art loaded");
		return;
	}
	i = 0;			/* Init index to ShapeBox */
	do {
		TextPtr = strtok(0,NumDelimiters);
		if (!TextPtr) {
			PrintError("Not enough parms for SLICE");
			return;
		}
		if (!isdigit(TextPtr[0]) && TextPtr[0]!='-') {
			PrintError("Non-numeric parm for SLICE");
			return;
		}
		ShapeBox[i] = AsciiToInteger(TextPtr);
		if (!i && ShapeBox[0]==(Word)-1) {	/* If parm #1 = -1 */
			ShapeBox[0] = 0;		/* Use the entire picture */
			ShapeBox[1] = 0;
			ShapeBox[2] = PicWidth;		/* Then use the shape's size as input */
			ShapeBox[3] = PicHeight;
			break;
		}
	} while (++i<4);			/* All 4 parms fetched? */

	AdjustX = 0;
	AdjustY = 0;
	TextPtr = strtok(0,NumDelimiters);
	if (TextPtr && (isdigit(TextPtr[0]) || TextPtr[0]=='-')) {
		AdjustX = AsciiToInteger(TextPtr);
		TextPtr = strtok(0,NumDelimiters);
		if (TextPtr && (isdigit(TextPtr[0]) || TextPtr[0]=='-')) {
			AdjustY = AsciiToInteger(TextPtr);
		}
	}

	if (!ShapeBox[2] || !ShapeBox[3]) {
		PrintError("Can't cut an empty box");
		return;
	}
	if ((ShapeBox[0] >= PicWidth) || (ShapeBox[1] >= PicHeight)) {
		PrintError("X/Y is out of bounds in the source pic");
	}
	i = 0;      /* Assume I don't clip */
	if ((ShapeBox[0] + ShapeBox[2]) > PicWidth) {
		ShapeBox[2] = PicWidth-ShapeBox[0];
		PrintError("Shape width was clipped");
	}
	if ((ShapeBox[1] + ShapeBox[3]) > PicHeight) {
		ShapeBox[3] = PicHeight-ShapeBox[1];
		PrintError("Shape height was clipped");
	}
	ShapeSize = ShapeBox[2]*ShapeBox[3];
	if (MaskEnable) {
		ShapeSize *= 2; /* Double the memory size for masked shape */
	}
	ShapePtr = static_cast<Word8*>(Alloc(ShapeSize));   /* Get the memory */
	if (!ShapePtr) {
		PrintError("Not enough memory for slice");
		return;
	}
	MyShapePtr = &Shapes[ShapeCount];
	memset(MyShapePtr,0,sizeof(MyShape_t));
	MyShapePtr->ShapePtr = ShapePtr;       /* Save the pointer to the shape */
	MyShapePtr->ShapeWidth = ShapeBox[2];
	MyShapePtr->ShapeHeight = ShapeBox[3];
	MyShapePtr->OriginalShapeWidth = ShapeBox[2];
	MyShapePtr->OriginalShapeHeight = ShapeBox[3];
	MyShapePtr->AdjustX = AdjustX;
	MyShapePtr->AdjustY = AdjustY;
	MyShapePtr->ShapePal = g_pPalette;		/* Palette pointer */
	ArtPtr = &g_pPictureBuffer[(ShapeBox[1]*PicWidth)+ShapeBox[0]];
	Height = ShapeBox[3];       /* Store my counters */
	Width = ShapeBox[2];

	if (!MaskEnable) {          /* Masking enabled? */
		do {
			i = 0;
			do {
				ShapePtr[i] = TransTable[ArtPtr[i]];
			} while (++i<Width);
			ShapePtr+=Width;
			ArtPtr += PicWidth;
		} while (--Height);
	} else {
		MaskPtr = static_cast<Word8*>(Alloc(ShapeSize));
		if (!MaskPtr) {
			Free(ShapePtr);
			PrintError("Not enough memory for slice");
			return;
		}
		MyShapePtr->MaskPtr = MaskPtr;
		do {
			i = 0;
			do {
				PicByte = TransTable[ArtPtr[i]];
				if (PicByte == MaskValue) {
					MaskPtr[i] = 0xff;      /* Don't mask it */
					ShapePtr[i] = 0;
				} else {
					MaskPtr[i] = 0x00;      /* Mask it off */
					ShapePtr[i] = (Word8)PicByte;
				}
			} while (++i<Width);
			ShapePtr+=Width;
			MaskPtr+=Width;
			ArtPtr += PicWidth;
		} while (--Height);
	}
	if (SidewaysFlag) {
		RotateShape(MyShapePtr);
	}
	if (!ReduceEnable) {
		MyShapePtr->ReduceX = 0x7FFF;
	} else {
		ShrinkShape(MyShapePtr);
	}
	++ShapeCount;           /* I have a shape */
	printf(".");
//	ShapeSize = ReadTick();
//	if ((ShapeSize-LastTick)>=3) {
//		LastTick = ShapeSize;
		fflush(stdout);
//	}
}


/***************************************

	Convert an 8 bit image into
	4 bit for IIGS images
	Note: All shapes MUST be an even number wide

***************************************/

static void MashShape(Word8 *ShapePtr,Word Width,Word Height)
{
	Word8 *DestPtr;
	Word ByteWidth;
	Word i,j;

	ByteWidth = (Width+1)/2;    /* Get the true width */
	if (!ByteWidth || !Height) {        /* Failsafe */
		return;
	}
	DestPtr = ShapePtr;     /* Init the dest ptr */
	do {
		i = 0;
		j = 0;
		do {
			DestPtr[j] = ShapePtr[i]<<4;    /* Get the first pixel */
			++i;            /* Skip */
			DestPtr[j] |= ShapePtr[i] & 0xf;
			++j;
		} while (++i<Width);
		DestPtr+=ByteWidth;
		ShapePtr+=Width;
	} while (--Height);
}

static void Mask255(MyShape_t *MyShapePtr)
{
	Word32 Size;
	Word8 *Dest,*Src;

	Size = MyShapePtr->ShapeWidth*MyShapePtr->ShapeHeight;

	if (!Size) {
		return;
	}
	Dest = MyShapePtr->MaskPtr;
	Src = MyShapePtr->ShapePtr;
	do {
		if (Dest[0]) {
			Src[0] = 255;
		}
		++Dest;
		++Src;
	} while (--Size);
}

/***************************************

	Write a single shape to disk

***************************************/

static void WriteShape(Word Index,FILE *fo)
{
	Word Width,Height;
	Word ByteWidth;
	Word32 Size,Size2;
	MyShape_t *MyShapePtr;

	MyShapePtr = &Shapes[Index];

	if (HeaderFlag && MyShapePtr->ReduceX != 0x7FFF) {
		Width = MyShapePtr->ReduceX+MyShapePtr->AdjustX;
		Height = MyShapePtr->ReduceY+MyShapePtr->AdjustY;
		switch (SaveType) {
		case PC3DODOOM:
			Width = ((MyShapePtr->OriginalShapeWidth+MyShapePtr->AdjustX)/2)-MyShapePtr->ReduceY-1;
			Height = ((MyShapePtr->OriginalShapeHeight+MyShapePtr->AdjustY)-5)-MyShapePtr->ReduceX;
		case PCMAC:
		case PC3DO:
			WriteBig(fo,(Word16)Width);     /* Big endian word size */
			WriteBig(fo,(Word16)Height);
			break;
		case PCIIGS:
			fputc(Width,fo);
			fputc(Height,fo);
			break;
		case PCIBM:
		case PCWIN95:
			WriteLittle(fo,(Word16)Width);		/* Little endian */
			WriteLittle(fo,(Word16)Height);
			break;
		}
	}

	Width = MyShapePtr->ShapeWidth;
	Height = MyShapePtr->ShapeHeight;

	Size = (Word32)Width * (Word32) Height;  /* Make SURE it's a 32 bit mul */
	switch (SaveType) {
	case PCMAC:
		if (HeaderFlag) {
			WriteBig(fo,(Word16)Width);     /* Big endian word size */
			WriteBig(fo,(Word16)Height);
		}
		if (MyShapePtr->MaskPtr) {
			Mask255(MyShapePtr);		/* Convert the mask color to 255 */
		}
		break;
	case PCIBM:
	case PCWIN95:
		if (HeaderFlag) {
			WriteLittle(fo,(Word16)Width);        /* Little endian word size */
			WriteLittle(fo,(Word16)Height);
		}
		break;
	case PCIIGS:
		ByteWidth = (Width+1)/2;    /* Get the width in bytes */
		Size = ByteWidth*Height;    /* Reset the size */
		if (HeaderFlag) {
			fputc(ByteWidth,fo);        /* Save the BYTE width */
			fputc(Height,fo);
			if (MyShapePtr->MaskPtr) {
				Size2 = Size+4;
				WriteLittle(fo,(Word16)Size2);
			}
		}
		MashShape(MyShapePtr->ShapePtr,Width,Height);   /* Convert to IIGS */
		if (MyShapePtr->MaskPtr) {
			MashShape(MyShapePtr->MaskPtr,Width,Height);    /* Convert to IIGS */
		}
		break;
	case PCSCALEIBM :
		SaveScaleShape(MyShapePtr,fo,TRUE);
		return;

	case PCSCALEMAC :
		SaveScaleShape(MyShapePtr,fo,FALSE);
		return;

	case PC3DO:
	case PC3DODOOM:
		Save3DOShape(MyShapePtr,fo);		/* Use a subroutine */
		return;						/* Exit NOW! */
	default :
		if (HeaderFlag) {
			fputc(Width,fo);        /* Save the byte x/y */
			fputc(Height,fo);
			if (MyShapePtr->MaskPtr) {
				Size2 = Size+4;
				WriteBig(fo,(Word16)Size2);
			}
		}
		break;
	}
	fwrite(MyShapePtr->ShapePtr,1,Size,fo);     /* Save the data */
}

/***************************************

	Save the shape to disk

***************************************/

static void SaveShapes(void)
{
	FILE *fo;       /* Output file referance */
	char *FileName; /* Name of file to save */
	Word Index; /* Pointer index */
	Word32 CurrentIndex; /* File output index */

	if (!ShapeCount) {  /* Any shapes cut out? */
		PrintError("No shapes cut");
		return;
	}
	FileName = strtok(0,Delimiters);        /* Get the operand */
	if (!FileName) {
		PrintError("No filename in SAVE");  /* No operand! */
		return;
	}
	fo = fopen(FileName,"wb");  /* Save to this file */
	if (!fo) {
		PrintError("Can't open output shape file"); /* File error! */
		return;
	}
	if (ShapeCount!=1 || IndexArrayFlag) {        /* Don't write a header */
		Index = 0;          /* Init the shape # */
		CurrentIndex = ShapeCount*4;    /* Init the save index */
		fwrite(g_pPalette,4,ShapeCount,fo);   /* Write a bogus header (Push EOF) */
		do {
			fseek(fo,Index*4,SEEK_SET); /* Write the file offset */
			switch(SaveType) {
			case PCMAC :
			case PC3DO :
			case PCSCALEMAC :
			case PC3DODOOM :
				WriteBig(fo,(Word32)CurrentIndex); /* Save as BIG ENDIAN! */
				break;
			default :
				WriteLittle(fo,(Word32)CurrentIndex);     /* Save as LITTLE ENDIAN */
				break;
			}
			fseek(fo,CurrentIndex,SEEK_SET);    /* Set to the end of file */
			WriteShape(Index,fo);           /* Save the shape */
			CurrentIndex = ftell(fo);       /* Where did it go? */
		} while (++Index<ShapeCount);       /* All done? */
	} else {
		WriteShape(0,fo);       /* Just write the shape data */
	}
	fclose(fo);     /* Close the file */

	do {        /* Release all the memory */
		MyShape_t *MyShapePtr;
		--ShapeCount;

		MyShapePtr = &Shapes[ShapeCount];
		Free(MyShapePtr->ShapePtr);    /* Release the shape memory */
		if (MyShapePtr->MaskPtr) {     /* Is there a mask here? */
			Free(MyShapePtr->MaskPtr); /* Release the mask memory */
			MyShapePtr->MaskPtr = 0;   /* Mark as freed */
		}
	} while (ShapeCount);   /* More? */
	ShapeCount=0;       /* No more shapes */
	InitPalette = TRUE;
}

/***************************************

	Execute the script tokens
	Input with the FILE* of the open script file
	I assume LineNum == 0.

***************************************/

static void Slicer(FILE* fp)
{
	char *TextPtr;  /* Pointer to string token */
	Word i; /* Index */

	InitTranspose();    /* Init the transpose table */
	InitPalette = TRUE;
	while (fgets(InputLine,sizeof(InputLine),fp)) { /* Get a string */
		++LineNum;          /* Adjust the line # */
		TextPtr = strtok(InputLine,Delimiters); /* Get the first token */
		if (!TextPtr) {
			continue;
		}
		i = 0;      /* Check for the first command */
		if (isalnum(TextPtr[0])) {  /* Comment? */
			do {
				if (!StringCaseCompare(TextPtr,Commands[i])) { /* Match? */
					switch (i) {        /* Execute the command */
					case 0:
						LogMachine();   /* Target machine */
					case 1:				/* Art type (Not used) */
						break;
					case 2:
						LoadArt();      /* Load in a picture file */
						break;
					case 3:
						SavePalette();  /* Write the palette out */
						break;
					case 4:
						SliceArt();     /* Cut out some art */
						break;
					case 5:
						SaveShapes();   /* Save the shape file */
						break;
					case 6:
						LogMask();      /* Get the mask setting */
						break;
					case 7:
						LogTranspose(); /* Set the color transpose */
						break;
					case 8:
						LogReduce();    /* Enable shape reduction */
						break;
					case 9:
						LogHeader();	/* Enable / Disable shape data header */
						break;
					case 10:			/* Enable / Disable 90 degree rotation */
						LogSideways();
						break;
					case 11:
						SliceTile();	/* Slice a tiled font */
						break;
					case 12:
						LogPacked();	/* Enable/disable packing */
						break;
					case 13:
						LogIndexArray();	/* Get the index array */
						break;
					case 14:
						SliceFont();	/* Slice a Burgerlib font */

					}
					break;      /* Don't parse anymore! */
				}
			} while (++i<CommandCount); /* Keep checking */
		}
		if (i==CommandCount) {      /* Didn't find it? */
			printf("# Command %s not implemented\n",TextPtr);
		}
	}
}

/***************************************

	Main dispatcher for the converter

***************************************/

static char Instructions[] = {
"# Copyright 1994-96,2014 by Rebecca Ann Heineman\n"
"# This program will extract art using a script\n"
"# Usage: Slicer Infile\n"
"# Script commands are as follows:\n"
"# MACHINE \"IBM\" | \"MAC\" | \"IIGS\" | \"SNES\" | \"3DO\" | \"SCALEIBM\" | \n"
"#          \"SCALEMAC\"\n"
"# LOAD inputfilename\n"
"# PALETTE outputfilename\n"
"# SLICE (X,Y,WIDTH,HEIGHT) | -1 [,XOffset,YOffset]\n"
"# SAVE outputfilename\n"
"# MASK OFF | indexcolornumber\n"
"# TRANSPOSE OFF | indexbefore,indexafter\n"
"# REDUCE OFF | indexcolornumber\n"
"# SHAPEHEADER OFF | ON\n"
"# SIDEWAYS OFF | ON\n"
"# SLICETILE X,Y,Width,Height,TilesPerLine,TotalTiles,outputfilename\n"
"# PACKED OFF | ON\n"
"# INDEXARRAY OFF | ON\n"
"# SLICEFONT X,Y,Height,FirstAsciiNumber,TotalTiles,outputfilename\n"
"# Numeric input can be 60,0x3C,$3C\n"
"# Accepts \\\\ and ; comments\n"
"# Acceptable filetypes TGA,TGC,BMP,RLE,LBM,PCX,APF,GIF\n"
#if defined(BURGER_MSDOS)
"# Supports long filenames under a Windows 95 shell but not under PURE dos"
#endif
};

int BURGER_ANSIAPI main(Word argc,const char **argv)
{
	ConsoleApp MyApp(argc,argv);

	ErrorFlag = FALSE;
	if (argc!=2) {          /* Gotta have input arguments */
		printf(Instructions);
		return 1;
	}

	FILE *fp = fopen(argv[1],"r");    /* Read the ASCII script */
	if (!fp) {
		printf("# Can't open script %s.\n",argv[1]);    /* Oh oh */
		return 1;
	}
	Palettes = (Word8 *)Alloc(256*768);
	Slicer(fp);     /* Slice it up! */
	Free(Palettes);		/* Release the memory */
	fclose(fp);     /* Close the script file */
	return ErrorFlag;		/* No error possible */
}
