#ifndef __SLICER_H__
#define __SLICER_H__

#ifndef __BURGER__
#include <Burger.h>
#endif

#include <stdio.h>

#define MAXSHAPES 512			/* Maximum number of shapes to slice in one time */

typedef struct CCB {
	Word32 ccb_Flags;
	Word32 ccb_NextPtr;
	Word32 ccb_SourcePtr;
	Word32 ccb_PLUTPtr;
	long ccb_XPos;
	long ccb_YPos;
	long ccb_HDX;
	long ccb_HDY;
	long ccb_VDX;
	long ccb_VDY;
	long ccb_HDDX;
	long ccb_HDDY;
	Word32 ccb_PIXC;
	Word32 ccb_PRE0;
	Word32 ccb_PRE1;
	long ccb_Width;			/* Width of shape in pixels */
	long ccb_Height;		/* Height of shape in pixels */
} CCB;

typedef struct {
	Word ShapeWidth;	/* Current shape width */
	Word ShapeHeight;	/* Current shape height */
	int ReduceX;		/* Left x clip value */
	int ReduceY;		/* Top y clip value */
	int AdjustX;		/* Extra adjustment x value */
	int AdjustY;		/* Extra adjustment y value */
	Word OriginalShapeWidth;	/* Original shape width before cropping */
	Word OriginalShapeHeight;	/* Original shape height before cropping */
	Word8 *ShapePtr;		/* Pointer to shape */
	Word8 *MaskPtr;		/* Pointer to mask */
	Word8 *ShapePal;		/* Pointer to palette */
} MyShape_t;

/* In Subs.c */

extern void PrintError(const char *Error);
extern void OffOn(const char *CommandName,Bool *Result);
extern void InitTranspose(void);

/* In Scale.c */

extern void SaveScaleShape(MyShape_t *MyShapePtr,FILE *fo,Word Little);

/* In Cut3DO.c */

extern Word ColorBits3DO;
extern void Save3DOShape(MyShape_t *MyShapePtr,FILE *fo);

/* In TileFont.c */

extern void SliceTile(void);

/* In CutFont.c */

extern void SliceFont(void);

/* In Reader.c */

extern Word ReadTGA(Word8 *Data);
extern Word ReadLBM(Word8 *Data);
extern Word ReadBMP(Word8 *Data);

/* In Data.c */

extern Word ErrorFlag;		/* Set to true if an error occured */
extern Word LineNum;		/* Input line # */
extern char Delimiters[];		/* Token delimiters */
extern char NumDelimiters[];	/* Value delimiters */
extern Bool PackedFlag;		/* True if shape data should be compressed */
extern Bool HeaderFlag;		/* True if no header should be written */
extern Bool SidewaysFlag;	/* True if the shape is rotated 90 degrees */
extern Bool MaskEnable;		/* Do I cut out with a mask? */
extern Bool IndexArrayFlag;	/* True if I always generate an index array */
extern MyShape_t Shapes[MAXSHAPES];	/* All clipped shapes */
extern Word8 Default3DOPal[128];
extern Word8 *g_pPictureBuffer;    /* Pointer to picture buffer */
extern Word8 *g_pPalette;  /* Extracted palette (256 RGBs) */
extern Word8 *Palettes;
extern Word PicHeight;     /* Height of picture in pixels */
extern Word PicWidth;      /* Width of picture in pixels */
extern Word MaskValue;         /* Mask value to use */
extern Word ShapeCount;        /* Number of shapes cut */
extern Word8 TransTable[256];  /* Shape color table */
extern Word ReduceColor;       /* Color to reduce by */
extern Word ReduceEnable;      /* Enable reduction? */

#endif
