/***************************************

	Global subroutines used by the slicer tool

***************************************/

#include "Slicer.h"
#include <String.h>
#include <Stdio.h>

/***************************************

	Print out a script error

***************************************/

void PrintError(const char *Error)
{
	ErrorFlag = 10;
	printf("# Error in Line %d, %s",LineNum,Error);
	WordPtr Length = StringLength(Error);
	if (Length) {
		--Length;
		if (Error[Length]!='\n') {
			putchar('\n');
		}
	}
}

/***************************************

	Fetch either a OFF or ON and set/clear the byte

***************************************/

void OffOn(const char *CommandName,Bool *Result)
{
	char ErrorTxt[100];
	char *TextPtr;

	TextPtr = strtok(0,NumDelimiters);
	if (!TextPtr) {
		sprintf(ErrorTxt,"No operand for %s",CommandName);
		PrintError(ErrorTxt);
		return;
	}
	if (!StringCaseCompare(TextPtr,"OFF")) {
		*Result = FALSE;
		return;
	}
	if (!StringCaseCompare(TextPtr,"ON")) {
		*Result = TRUE;
		return;
	}
	PrintError("Operand must be OFF or ON.");
}

/***************************************

	Init the color transposition table

***************************************/

void InitTranspose(void)
{
	Word i;
	i=0;
	do {
		TransTable[i] = static_cast<Word8>(i);  /* Use the numbers as is */
	} while (++i<256);
}


