#include "Slicer.h"
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

/***************************************

	Cut out a tileset, the tileset is only in the form of...

	SLICETILE FontX,FontY,CelWidth,CelHeight,CelCountPerLine,TotalCels,
		FileName

	There is no header saved, it will create a monospaced tile font
	with no header or palette information. This is for simple Ultima type
	fonts for limited use.

***************************************/

void SliceTile(void)
{
	Word ShapeBox[6];	/* Rect for the current shape */
	Word8 *ShapePtr;		/* Pointer to new shape */
	char *TextPtr;		/* Pointer to the slice index */
	Word i;             /* Temp */
	Word Width;         /* Width of cut */
	Word32 ShapeSize;	/* Size of the new font file */
	Word BaseX;
	Word Count;
	Word8 *ShapePtr2;

	if (!g_pPictureBuffer) {			/* Art loaded? */
		PrintError("No source art loaded");
		return;
	}

	i = 0;			/* Init index to ShapeBox */
	do {
		TextPtr = strtok(0,NumDelimiters);
		if (!TextPtr) {
			PrintError("Not enough parms for SLICETILE");
			return;
		}
		if (!isdigit(TextPtr[0])) {		/* First char must be a digit */
			PrintError("Non-numeric parm for SLICETILE");
			return;
		}
		ShapeBox[i] = AsciiToInteger(TextPtr);		/* Get the number */
	} while (++i<6);			/* All 6 parms fetched? */
	TextPtr = strtok(0,NumDelimiters);	/* Get the filename */

	if (!ShapeBox[2] || !ShapeBox[3]) {
		PrintError("Can't cut an empty box");
		return;
	}
	ShapeSize = ShapeBox[2]*ShapeBox[3]*ShapeBox[5];	/* Size of font */
	ShapePtr = (Word8 *)Alloc(ShapeSize);   /* Get the memory */

	if (!ShapePtr) {
		PrintError("Not enough memory for SLICETILE");
		return;
	}
	BaseX = ShapeBox[0];	/* Init the X coord */
	Count = ShapeBox[4];	/* Cels per line */
	Width = ShapeBox[2];	/* Width of cel */
	if (!Count) {
		PrintError("Cels per line is zero");
		Free(ShapePtr);
		return;
	}
	if (!ShapeBox[5]) {
		PrintError("Total Cels are zero");
		Free(ShapePtr);
		return;
	}

	ShapePtr2 = ShapePtr;
	do {
		Word8 *ArtPtr;		/* Pointer to the source art */
		Word Height;		/* Height of the shape to cut */
		ArtPtr = &g_pPictureBuffer[(ShapeBox[1]*PicWidth)+BaseX];
		Height = ShapeBox[3];       /* Store my counters */
		do {
			i = 0;
			do {
				ShapePtr2[i] = TransTable[ArtPtr[i]];
			} while (++i<Width);
			ShapePtr2+=Width;
			ArtPtr += PicWidth;
		} while (--Height);
		BaseX+=ShapeBox[2];	/* Move the X coord */
		if (!--Count) {		/* Next line down? */
			BaseX = ShapeBox[0];	/* Reset the X coord */
			ShapeBox[1]+=ShapeBox[3];	/* Move the Y coord down one line */
			Count = ShapeBox[4];	/* Reset the count */
		}
	} while (--ShapeBox[5]);

	FileManager::SaveFile(TextPtr,ShapePtr,ShapeSize);	/* Save the output */
	Free(ShapePtr);		/* Release the buffer */
}

